@@ -52,7 +52,7 @@ int
 QPDFAnnotationObjectHelper::getFlags()
 {
     QPDFObjectHandle flags_obj = this->oh.getKey("/F");
-    return flags_obj.isInteger() ? flags_obj.getIntValue() : 0;
+    return flags_obj.isInteger() ? flags_obj.getIntValueAsInt() : 0;
 }
 
 QPDFObjectHandle
